==> helpdesk_zendesk-1.0.13.tar.gz: helpdesk_zendesk/src/Plugin/HelpdeskIntegration/Zendesk.php <==
<?php

namespace Drupal\helpdesk_zendesk\Plugin\HelpdeskIntegration;

use Drupal\comment\CommentInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\file\Entity\File;
use Drupal\helpdesk_integration\Entity\Issue;
use Drupal\helpdesk_integration\HelpdeskInterface;
use Drupal\helpdesk_integration\HelpdeskPluginException;
use Drupal\helpdesk_integration\IssueInterface;
use Drupal\helpdesk_integration\PluginBase;
use Drupal\migrate\Audit\AuditException;
use Drupal\user\UserInterface;
use Exception;
use Zendesk\API\Exceptions\ApiResponseException;
use Zendesk\API\Exceptions\AuthException;
use Zendesk\API\Exceptions\CustomException;
use Zendesk\API\Exceptions\MissingParametersException;
use Zendesk\API\Exceptions\ResponseException;
use Zendesk\API\HttpClient;
use function PHPUnit\Framework\isEmpty;

/**
 * Plugin implementation of the Zendesk helpdesk.
 *
 * @HelpdeskPlugin(
 *   id = "zendesk",
 *   label = @Translation("Zendesk"),
 *   description = @Translation("Provides the Zendesk helpdesk plugin.")
 * )
 */
class Zendesk extends PluginBase {

  const TICKET_RESTRICTION = 'organization';

  use StringTranslationTrait;

  /**
   * {@inheritdoc}
   */
  public static function settingKeys(): array {
    return [
      'subdomain',
      'admin_user',
      'api_token',
      'group',
      'max_file_size',
      'states',
      'state_closed',
    ];
  }

  /**
   * {@inheritdoc}
   * @throws \Drupal\helpdesk_integration\HelpdeskPluginException
   */
  public function buildConfigurationForm(HelpdeskInterface $helpdesk, array $required): array {
    $form['subdomain'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Subdomain'),
        '#default_value' => $helpdesk->get('subdomain'),
      ] + $required;
    $form['admin_user'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Admin User'),
        '#default_value' => $helpdesk->get('admin_user'),
      ] + $required;
    $form['api_token'] = [
        '#type' => 'textfield',
        '#title' => $this->t('API Token'),
        '#default_value' => $helpdesk->get('api_token'),
      ] + $required;
    $form['max_file_size'] = [
        '#type' => 'number',
        '#title' => $this->t('Max. file size for attachments (in MB)'),
        '#default_value' => $helpdesk->get('max_file_size'),
        '#min' => 1,
      ] + $required;
    $states = $helpdesk->get('subdomain') ? $this->getStates($helpdesk) : [];
    $form['states'] = [
      '#type' => 'value',
      '#value' => $states,
    ];
    $form['state_closed'] = [
      '#type' => 'select',
      '#title' => $this->t('Closed ticket state'),
      '#default_value' => $helpdesk->get('state_closed'),
      '#options' => $states,
    ];

    return $form;
  }

  /**
   * Gets the API Client.
   *
   * @param \Drupal\helpdesk_integration\HelpdeskInterface $helpdesk
   *
   * @return \Zendesk\API\HttpClient
   * @throws \Drupal\helpdesk_integration\HelpdeskPluginException
   */
  public static function getClient(HelpdeskInterface $helpdesk, string $email = NULL): HttpClient {
    $subdomain = $helpdesk->get('subdomain');
    $userName = $helpdesk->get('admin_user');
    if ($email) {
      $userName = $email;
    }
    $token = $helpdesk->get('api_token');
    $client = new HttpClient($subdomain);
    try {
      $client->setAuth('basic', ['username' => $userName, 'token' => $token]);
    } catch (AuthException $e) {
      throw new HelpdeskPluginException($e);
    }

    return $client;
  }

  /**
   * Get all states from the helpdesk.
   *
   * @param \Drupal\helpdesk_integration\HelpdeskInterface $helpdesk
   *   The helpdesk entity with the instance's configuration.
   *
   * @return array
   *   List of all active states indexed by their remote state ID.
   * @throws \Drupal\helpdesk_integration\HelpdeskPluginException
   */
  private function getStates(HelpdeskInterface $helpdesk): array {
    $states = [];
    try {
      $client = self::getClient($helpdesk);
      foreach ($client->ticketFields()->findAll()->ticket_fields as $state) {
        if ($state->type === 'status') {
          foreach ($state->system_field_options as $option) {
            $states[$option->value] = $option->name;
          }
        }
      }
    } catch (Exception $e) {
      $this->logger->error($e->getMessage());
    }
    return $states;
  }

  /**
   * {@inheritdoc}
   */
  public function createIssue(HelpdeskInterface $helpdesk, IssueInterface $issue) {
    try {
      $client = self::getClient($helpdesk, $issue->getOwner()->getEmail());
      $uploadTokens = $this->getUploadTokens($client, $issue->get('attachments')
        ->getValue());
      $bodyValue = $issue->get('body')->value;
      if(!$bodyValue) {
        $bodyValue = 'no text';
      }
      $remote_issue = $client->requests()->create([
        'subject' => $issue->label(),
        'comment' => [
          'html_body' => $bodyValue,
          'uploads' => $uploadTokens,
        ],
      ]);
      $issue->set('extid', $remote_issue->request->id);

    } catch (Exception $e) {
      $errorMessage = $e->getMessage();
      if ($e instanceof ApiResponseException && $e->getCode() === 422) {
        $errorDetails = json_decode($e->getErrorDetails());
        if ($baseError = $errorDetails->details->base) {
          $errorMessage = $baseError[0]->description;
        } else if ( $baseError = $errorDetails->details->email) {
          $errorMessage = $baseError[0]->description;
        }
      }
      throw new HelpdeskPluginException($errorMessage);
    }
  }

  /**
   * Get all upload tokens for the attachments.
   *
   * @throws \Zendesk\API\Exceptions\MissingParametersException
   * @throws \Zendesk\API\Exceptions\CustomException
   */
  function getUploadTokens(HttpClient $client, array $attachments): array {
    $uploadTokens = [];
    /** @var \Drupal\file\Plugin\Field\FieldType\FileItem $attachment */
    foreach ($attachments as $attachment) {
      $file = File::load($attachment['target_id']);
      $attachmentData = [
        'file' => $file->getFileUri(),
        'type' => $file->getMimeType(),
        'name' => $file->getFilename(),
      ];

      $uploadTokens[] = $client->attachments()
        ->upload($attachmentData)->upload->token;
    }
    return $uploadTokens;
  }

  /**
   * {@inheritdoc}
   */
  public function addCommentToIssue(HelpdeskInterface $helpdesk, IssueInterface $issue, CommentInterface $comment) {
    try {
      $client = self::getClient($helpdesk, $issue->getOwner()->getEmail());
      $ticket_id = $issue->get('extid')->value;
      $uploadTokens = $this->getUploadTokens($client, $comment->get('field_attachments')
        ->getValue());
      $client->requests()->update($ticket_id, [
        'comment' => [
          'html_body' => $comment->get('comment_body')->value,
          'uploads' => $uploadTokens,
        ],
      ]);


      $remote_comments = $client->requests($ticket_id)
        ->comments()
        ->findAll()->comments;
      foreach ($remote_comments as $remote_comment) {
        $comment->set('field_extid', $remote_comment->id);
      }
    } catch (ApiResponseException|AuthException|CustomException|MissingParametersException  $e) {
      throw new HelpdeskPluginException($e);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function resolveIssue(HelpdeskInterface $helpdesk, IssueInterface $issue): void {
    $closed_state = $helpdesk->get('state_closed');
    $client = self::getClient($helpdesk);

    try {
      $client->tickets()->update($issue->get('extid')->value, [
        'status' => $closed_state,
      ]);
    } catch (Exception $exception) {
      $errorMessage = $exception->getMessage();
      if ($exception instanceof ApiResponseException && $exception->getCode() === 422) {
        $errorDetails = json_decode($exception->getErrorDetails());
        if ($emailError = $errorDetails->details->base) {
          $errorMessage = $emailError[0]->description;
        }
      }
      throw new HelpdeskPluginException($errorMessage);
    }

    $states = $helpdesk->get('states');
    $issue->set('status', $states[$closed_state] ?? '');
  }

  /**
   * {@inheritdoc}
   */
  public function getAllIssues(HelpdeskInterface $helpdesk, UserInterface $user, int $since = 0): array {
    $states = $helpdesk->get('states');
    $state_closed = (string) $helpdesk->get('state_closed');
    $client = self::getClient($helpdesk);
    $issues = [];
    try {
      // NOTE: We do the update date check programmatically, because Zendesk
      // needs a few  minutes to update the search indexes.
      // See https://support.zendesk.com/hc/en-us/articles/203663226
      $requests = $client
        ->users($this->getUserData($user, $helpdesk, self::REMOTE_ID))
        ->requests()
        ->findAll()->requests;
      foreach ($requests as $request) {
        /** @var \Drupal\helpdesk_integration\IssueInterface $issue */
        $issue = Issue::create([
          'helpdesk' => $helpdesk->id(),
          'extid' => $request->id,
          'resolved' => (string) $request->status === $state_closed,
          'title' => $request->subject,
          'status' => $states[$request->status] ?? '',
          'body' => [
            'value' => empty($request->description) ? '' : $request->description,
            'format' => 'basic_html',
          ],
          'created' => strtotime($request->created_at),
          'changed' => strtotime($request->updated_at),
        ]);

        $this->handleComments($client, $request, $issue, $user);
        $issues[] = $issue;
      }
    } catch (ApiResponseException|AuditException|AuthException $e) {
      throw new HelpdeskPluginException($e);
    }
    return $issues;
  }

  /**
   * Handle ticket comments. The first comment is special and must be handled
   * differently. The attachments, if present,  are stored in the issue itself.
   *
   * @param \Zendesk\API\HttpClient $client
   * @param object $ticket
   * @param \Drupal\helpdesk_integration\IssueInterface $issue
   * @param \Drupal\user\UserInterface $user
   *
   * @throws \Zendesk\API\Exceptions\ApiResponseException
   * @throws \Zendesk\API\Exceptions\AuthException
   */
  private function handleComments(HttpClient $client, object $ticket, IssueInterface $issue, UserInterface $user): void {
    $count = 0;
    $comments = $client->requests($ticket->id)->comments()->findAll()->comments;
    foreach ($comments as $comment) {
      if ($count > 0) {
        $comment_id = $issue->addComment(
          $comment->id,
          $comment->body,
          $user,
          strtotime($comment->created_at),
          strtotime($comment->created_at)
        );
        $this->handleCommentAttachments($comment, $issue, $comment_id);
      }
      else {
        foreach ($comment->attachments as $attachment) {
          $issue->addIssueAttachment($attachment->file_name, $attachment->content_url);
        }
      }
      $count++;
    }
  }

  /**
   * Handles the attachments of the ticket comments.
   *
   * @param object $comment
   * @param \Drupal\helpdesk_integration\IssueInterface $issue
   * @param int $comment_id
   */
  private function handleCommentAttachments(object $comment, IssueInterface $issue, int $comment_id): void {
    foreach ($comment->attachments as $attachment) {
      $issue->addCommentAttachment(
        $comment_id,
        $attachment->file_name,
        $attachment->content_url
      );
    }
  }

  /**
   * {@inheritdoc}
   */
  public function pushUser(HelpdeskInterface $helpdesk, UserInterface $user): string {
    $client = self::getClient($helpdesk);
    try {
      if ($remote_id = $this->getUserData($user, $helpdesk, self::REMOTE_ID)) {
        $remoteUser = $client->users()->find($remote_id);
        if ($remoteUser->user->active) {
          $this->updateUser($client, $remoteUser->user, $user);
          return $remote_id;
        }
        else {
          $userFromPlatform = $this->createUser($client, $user);
          return $userFromPlatform->user->id;
        }
      }
      else {
        $remote_users = $client->users()
          ->search(['query' => $user->getEmail()]);
        if (!empty($remote_users->users)) {
          $this->updateUser($client, $remote_users->users[0], $user);
          return $remote_users->users[0]->id;
        }
        else {
          $userFromPlatform = $this->createUser($client, $user);
          return $userFromPlatform->user->id;
        }

      }
    } catch (Exception $e) {
      $errorMessage = $e->getMessage();
      if ($e instanceof ApiResponseException && $e->getCode() === 422) {
        $errorDetails = json_decode($e->getErrorDetails());
        if ($emailError = $errorDetails->details->email) {
          $errorMessage = $emailError[0]->description;
        }
      }
      throw new HelpdeskPluginException($errorMessage);
    }
  }

  /**
   * Updates the user.
   *
   * @param \Zendesk\API\HttpClient $client
   * @param object $remoteUser
   * @param \Drupal\user\UserInterface $user
   *
   * @throws \Zendesk\API\Exceptions\ApiResponseException
   * @throws \Zendesk\API\Exceptions\AuthException
   * @throws \Zendesk\API\Exceptions\MissingParametersException
   */
  private function updateUser(HttpClient $client, object $remoteUser, UserInterface $user): void {
    $hasChanged = FALSE;
    if(strcasecmp($remoteUser->email, $user->getEmail()) !== 0) {
      $hasChanged = TRUE;
    }
    if($remoteUser->name != $user->getDisplayName()) {
      $hasChanged = TRUE;
    }
    if($hasChanged) {
      $client->users()->update($remoteUser->id, $this->getUserArray($user));
      $this->deleteAllOldEmailIdentities($client, $remoteUser->id, $user->getEmail());
    }
  }

  /**
   * Gets the email identities of a Zendesk user.
   *
   * @param \Zendesk\API\HttpClient $client
   * @param string $remoteUserId
   *
   * @return array
   * @throws \Zendesk\API\Exceptions\ApiResponseException
   * @throws \Zendesk\API\Exceptions\AuthException
   */
  private function getUserEmailIdentities(HttpClient $client, string $remoteUserId): array {
    return $this->getUserIdentities($client, $remoteUserId, 'email');
  }

  /**
   * Gets the user identities by type.
   *
   * @param \Zendesk\API\HttpClient $client
   * @param string $remoteUserId
   * @param string $identityType
   *
   * @return array
   * @throws \Zendesk\API\Exceptions\ApiResponseException
   * @throws \Zendesk\API\Exceptions\AuthException
   */
  private function getUserIdentities(HttpClient $client, string $remoteUserId, string $identityType): array {
    $identities = [];
    foreach ($client->users($remoteUserId)
               ->identities()
               ->findAll()->identities as $identity) {
      if ($identity->type == $identityType) {
        $identities[] = $identity;
      }
    }

    return $identities;
  }

  /**
   * Delete all old email identities, but not the current one.
   *
   * @param \Zendesk\API\HttpClient $client
   * @param string $remoteUserId
   * @param string $currentEmail
   *
   * @throws \Zendesk\API\Exceptions\ApiResponseException
   * @throws \Zendesk\API\Exceptions\AuthException
   * @throws \Zendesk\API\Exceptions\MissingParametersException
   */
  private function deleteAllOldEmailIdentities(HttpClient $client, string $remoteUserId, string $currentEmail): void {
    foreach ($this->getUserEmailIdentities($client, $remoteUserId) as $identity) {
      if(strcasecmp($currentEmail, $identity->value) === 0) {
        continue;
      }
      $client->users($remoteUserId)->identities()->delete($identity->id);
    }
  }

  /**
   * Create the user.
   *
   * @param \Zendesk\API\HttpClient $client
   * @param \Drupal\user\UserInterface $user
   *
   * @return object|null
   */
  private function createUser(HttpClient $client, UserInterface $user): ?object {
    return $client->users()->create($this->getUserArray($user));
  }

  /**
   * Gets the values of the user as array.
   *
   * @param \Drupal\user\UserInterface $user
   *
   * @return array
   */
  private function getUserArray(UserInterface $user): array {
    return [
      'name' => $user->getDisplayName(),
      'email' => $user->getEmail(),
      'verified' => TRUE,
      'ticket_restriction' => self::TICKET_RESTRICTION,
    ];
  }

  /**
   * @inheritDoc
   */
  public function isUserLocked(HelpdeskInterface $helpdesk, UserInterface $user): bool {
    $client = self::getClient($helpdesk);
    try {
      if ($remote_id = $this->getUserData($user, $helpdesk, self::REMOTE_ID)) {
        $remoteUser = $client->users()->find($remote_id);
        if ($remoteUser->user->suspended) {
          return TRUE;
        }
      }
    } catch (MissingParametersException $e) {
      throw new HelpdeskPluginException($e);
    }
    return FALSE;
  }

}

==> helpdesk_zendesk-1.0.15.tar.gz: helpdesk_zendesk/src/Plugin/HelpdeskIntegration/Zendesk.php <==
<?php

namespace Drupal\helpdesk_zendesk\Plugin\HelpdeskIntegration;

use Drupal\comment\CommentInterface;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Http\ClientFactory;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\helpdesk_integration\Entity\Issue;
use Drupal\helpdesk_integration\HelpdeskInterface;
use Drupal\helpdesk_integration\HelpdeskPluginException;
use Drupal\helpdesk_integration\IssueInterface;
use Drupal\helpdesk_integration\PluginBase;
use Drupal\helpdesk_zendesk\Service;
use Drupal\migrate\Audit\AuditException;
use Drupal\user\UserDataInterface;
use Drupal\user\UserInterface;
use Exception;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Mime\MimeTypeGuesserInterface;
use Zendesk\API\Exceptions\ApiResponseException;
use Zendesk\API\Exceptions\AuthException;
use Zendesk\API\Exceptions\MissingParametersException;
use Zendesk\API\HttpClient;

/**
 * Plugin implementation of the Zendesk helpdesk.
 *
 * @HelpdeskPlugin(
 *   id = "zendesk",
 *   label = @Translation("Zendesk"),
 *   description = @Translation("Provides the Zendesk helpdesk plugin.")
 * )
 */
class Zendesk extends PluginBase {

  use StringTranslationTrait;

  /**
   * The Zendesk service.
   *
   * @var Service
   */
  protected $zendeskService;

  /**
   * Constructor.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Http\ClientFactory $client_factory
   *   HTTP client factory.
   * @param \Drupal\helpdesk_integration\Service $service
   *   Helpdesk services.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   Entity type manager service.
   * @param \Drupal\user\UserDataInterface $user_data
   *   User data service to persistently store data for each user.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   Drupal's time service.
   * @param \Drupal\Core\Datetime\DateFormatterInterface $date_formatter
   *   Drupal's date formatter service.
   * @param \Psr\Log\LoggerInterface $logger
   *   Drupal's logger service.
   * @param \Drupal\Core\Session\AccountInterface $current_user
   *   The currently logged in user account session.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system interface.
   * @param \Symfony\Component\Mime\MimeTypeGuesserInterface $mime_type_guesser
   *   The file type guesser.
   * @param \Drupal\helpdesk_zendesk\Service $zendeskService
   *   The Zendesk service.
   */
  public function __construct(array $configuration, string $plugin_id, $plugin_definition, ClientFactory $client_factory, \Drupal\helpdesk_integration\Service $service, EntityTypeManagerInterface $entity_type_manager, UserDataInterface $user_data, TimeInterface $time, DateFormatterInterface $date_formatter, LoggerInterface $logger, AccountInterface $current_user, FileSystemInterface $file_system, MimeTypeGuesserInterface $mime_type_guesser, Service $zendeskService) {
    parent::__construct($configuration,$plugin_id,$plugin_definition,$client_factory,$service,$entity_type_manager,$user_data,$time,$date_formatter,$logger,$current_user,$file_system,$mime_type_guesser);
    $this->zendeskService = $zendeskService;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('http_client_factory'),
      $container->get('helpdesk_integration.service'),
      $container->get('entity_type.manager'),
      $container->get('user.data'),
      $container->get('datetime.time'),
      $container->get('date.formatter'),
      $container->get('logger.factory')->get('helpdesk'),
      $container->get('current_user'),
      $container->get('file_system'),
      $container->get('file.mime_type.guesser'),
      $container->get('helpdesk_zendesk.service'),
    );

  }

  /**
   * {@inheritdoc}
   */
  public static function settingKeys(): array {
    return [
      'subdomain',
      'admin_user',
      'api_token',
      'group',
      'max_file_size',
      'states',
      'state_closed',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(HelpdeskInterface $helpdesk, array $required): array {
    $form['subdomain'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Subdomain'),
        '#default_value' => $helpdesk->get('subdomain'),
      ] + $required;
    $form['admin_user'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Admin User'),
        '#default_value' => $helpdesk->get('admin_user'),
      ] + $required;
    $form['api_token'] = [
        '#type' => 'textfield',
        '#title' => $this->t('API Token'),
        '#default_value' => $helpdesk->get('api_token'),
      ] + $required;
    $form['max_file_size'] = [
        '#type' => 'number',
        '#title' => $this->t('Max. file size for attachments (in MB)'),
        '#default_value' => $helpdesk->get('max_file_size'),
        '#min' => 1,
      ] + $required;
    $states = $helpdesk->get('subdomain') ? $this->getStates($helpdesk) : [];
    $form['states'] = [
      '#type' => 'value',
      '#value' => $states,
    ];
    $form['state_closed'] = [
      '#type' => 'select',
      '#title' => $this->t('Closed ticket state'),
      '#default_value' => $helpdesk->get('state_closed'),
      '#options' => $states,
    ];

    return $form;
  }

  /**
   * Get all states from the helpdesk.
   *
   * @param HelpdeskInterface $helpdesk
   *   The helpdesk entity with the instance's configuration.
   *
   * @return array
   *   List of all active states indexed by their remote state ID.
   */
  private function getStates(HelpdeskInterface $helpdesk): array {
    $states = [];
    try {
      $client = $this->zendeskService->getClient($helpdesk);
      $states = $this->zendeskService->getStates($client);
    } catch (Exception $e) {
      $this->logger->error($e->getMessage());
    }
    return $states;
  }

  /**
   * {@inheritdoc}
   */
  public function createIssue(HelpdeskInterface $helpdesk, IssueInterface $issue) {
    try {
      $client = $this->zendeskService->getClient($helpdesk, $issue->getOwner()->getEmail());
      $uploadTokens = $this->zendeskService->uploadTokens($client, $issue->get('attachments')
        ->getValue());
      $bodyValue = $issue->get('body')->value;
      if (!$bodyValue) {
        $bodyValue = 'no text';
      }
      $remote_issue = $this->zendeskService->createIssue($client, $issue->label(), $bodyValue, $uploadTokens);
      $issue->set('extid', $remote_issue->request->id);

    } catch (Exception $e) {
      $errorMessage = $e->getMessage();
      if ($e instanceof ApiResponseException && $e->getCode() === 422) {
        $errorDetails = json_decode($e->getErrorDetails());
        if ($baseError = $errorDetails->details->base) {
          $errorMessage = $baseError[0]->description;
        }
        else {
          if ($baseError = $errorDetails->details->email) {
            $errorMessage = $baseError[0]->description;
          }
        }
      }
      throw new HelpdeskPluginException($errorMessage);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function addCommentToIssue(HelpdeskInterface $helpdesk, IssueInterface $issue, CommentInterface $comment) {
    try {
      $client = $this->zendeskService->getClient($helpdesk, $issue->getOwner()->getEmail());
      $ticket_id = $issue->get('extid')->value;
      $uploadTokens = $this->zendeskService->uploadTokens($client, $comment->get('field_attachments')
        ->getValue());
      $client->requests()->update($ticket_id, [
        'comment' => [
          'html_body' => $comment->get('comment_body')->value,
          'uploads' => $uploadTokens,
        ],
      ]);

      $remote_comments = $client->requests($ticket_id)
        ->comments()
        ->findAll()->comments;
      foreach ($remote_comments as $remote_comment) {
        $comment->set('field_extid', $remote_comment->id);
      }
    } catch (ApiResponseException|AuthException  $e) {
      throw new HelpdeskPluginException($e);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function resolveIssue(HelpdeskInterface $helpdesk, IssueInterface $issue): void {
    $closed_state = $helpdesk->get('state_closed');
    $client = $this->zendeskService->getClient($helpdesk);

    try {
      $client->tickets()->update($issue->get('extid')->value, [
        'status' => $closed_state,
      ]);
    } catch (Exception $exception) {
      $errorMessage = $exception->getMessage();
      if ($exception instanceof ApiResponseException && $exception->getCode() === 422) {
        $errorDetails = json_decode($exception->getErrorDetails());
        if ($emailError = $errorDetails->details->base) {
          $errorMessage = $emailError[0]->description;
        }
      }
      throw new HelpdeskPluginException($errorMessage);
    }

    $states = $helpdesk->get('states');
    $issue->set('status', $states[$closed_state] ?? '');
  }

  /**
   * {@inheritdoc}
   */
  public function getAllIssues(HelpdeskInterface $helpdesk, UserInterface $user, int $since = 0): array {
    $states = $helpdesk->get('states');
    $state_closed = (string) $helpdesk->get('state_closed');
    $client = $this->zendeskService->getClient($helpdesk);
    $issues = [];
    try {
      // NOTE: We do the update date check programmatically, because Zendesk
      // needs a few  minutes to update the search indexes.
      // See https://support.zendesk.com/hc/en-us/articles/203663226
      $requests = $client
        ->users($this->getUserData($user, $helpdesk, self::REMOTE_ID))
        ->requests()
        ->findAll()->requests;
      foreach ($requests as $request) {
        /** @var \Drupal\helpdesk_integration\IssueInterface $issue */
        $issue = Issue::create([
          'helpdesk' => $helpdesk->id(),
          'extid' => $request->id,
          'resolved' => (string) $request->status === $state_closed,
          'title' => $request->subject,
          'status' => $states[$request->status] ?? '',
          'body' => [
            'value' => empty($request->description) ? '' : $request->description,
            'format' => 'basic_html',
          ],
          'created' => strtotime($request->created_at),
          'changed' => strtotime($request->updated_at),
        ]);

        $this->handleComments($client, $request, $issue, $user);
        $issues[] = $issue;
      }
    } catch (ApiResponseException|AuditException|AuthException $e) {
      throw new HelpdeskPluginException($e);
    }
    return $issues;
  }

  /**
   * Handle ticket comments. The first comment is special and must be handled
   * differently. The attachments, if present,  are stored in the issue itself.
   *
   * @param HttpClient $client
   * @param object $ticket
   * @param IssueInterface $issue
   * @param UserInterface $user
   *
   * @throws ApiResponseException
   * @throws AuthException
   */
  private function handleComments(HttpClient $client, object $ticket, IssueInterface $issue, UserInterface $user): void {
    $count = 0;
    $comments = $client->requests($ticket->id)->comments()->findAll()->comments;
    foreach ($comments as $comment) {
      if ($count > 0) {
        $comment_id = $issue->addComment(
          $comment->id,
          $comment->body,
          $user,
          strtotime($comment->created_at),
          strtotime($comment->created_at)
        );
        $this->handleCommentAttachments($comment, $issue, $comment_id);
      }
      else {
        foreach ($comment->attachments as $attachment) {
          $issue->addIssueAttachment($attachment->file_name, $attachment->content_url);
        }
      }
      $count++;
    }
  }

  /**
   * Handles the attachments of the ticket comments.
   *
   * @param object $comment
   * @param \Drupal\helpdesk_integration\IssueInterface $issue
   * @param int $comment_id
   */
  private function handleCommentAttachments(object $comment, IssueInterface $issue, int $comment_id): void {
    foreach ($comment->attachments as $attachment) {
      $issue->addCommentAttachment(
        $comment_id,
        $attachment->file_name,
        $attachment->content_url
      );
    }
  }

  /**
   * {@inheritdoc}
   */
  public function pushUser(HelpdeskInterface $helpdesk, UserInterface $user): string {
    $client = $this->zendeskService->getClient($helpdesk);
    try {
      if ($remote_id = $this->getUserData($user, $helpdesk, self::REMOTE_ID)) {
        $remoteUser = $client->users()->find($remote_id);
        if ($remoteUser->user->active) {
          $this->zendeskService->updateUser($client, $remoteUser->user, $user);
          return $remote_id;
        }
        else {
          $userFromPlatform = $this->zendeskService->createUser($client, $user);
          return $userFromPlatform->user->id;
        }
      }
      else {
        $remote_users = $client->users()
          ->search(['query' => $user->getEmail()]);
        if (!empty($remote_users->users)) {
          $this->zendeskService->updateUser($client, $remote_users->users[0], $user);
          return $remote_users->users[0]->id;
        }
        else {
          $userFromPlatform = $this->zendeskService->createUser($client, $user);
          return $userFromPlatform->user->id;
        }

      }
    } catch (Exception $e) {
      $errorMessage = $e->getMessage();
      if ($e instanceof ApiResponseException && $e->getCode() === 422) {
        $errorDetails = json_decode($e->getErrorDetails());
        if ($emailError = $errorDetails->details->email) {
          $errorMessage = $emailError[0]->description;
        }
      }
      throw new HelpdeskPluginException($errorMessage);
    }
  }

  /**
   * @inheritDoc
   */
  public function isUserLocked(HelpdeskInterface $helpdesk, UserInterface $user): bool {
    $client = $this->zendeskService->getClient($helpdesk);
    try {
      if ($remote_id = $this->getUserData($user, $helpdesk, self::REMOTE_ID)) {
        $remoteUser = $client->users()->find($remote_id);
        if ($remoteUser->user->suspended) {
          return TRUE;
        }
      }
    } catch (MissingParametersException $e) {
      throw new HelpdeskPluginException($e);
    }
    return FALSE;
  }

}
